/*
 * key-parse.c -- parse a key-list file and decode a stroke using that info
 *
 * Copyright (c) 2000 Alessandro Rubini (rubini@gnu.org)
 * Sponsored by Linuxcare Italia Spa.
 *
 * Extensive modifications by Jorj Bauer <jorj@seas.upenn.edu>,
 * sponsored by DejaVu Software, Inc. (http://www.dejavusoftware.com).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h> /* for atan2 */

#include "key-engine.h"


/* some internal constants */
#define LINE_LEN 128

/* Unfortunately, I'm having great problems with this */
//#define USE_STROKE_TOO /* undef to use landing-point-only behaviour */

/* I tried to use 8 directions, but it doesn't prove stable */
#define STROKE_NONE      0

#define STROKE_SPACE     1 /* east */
                           /* north-east */
#define STROKE_SHIFT     3 /* north */
                           /* north-west */
#define STROKE_DEL       5 /* west */
#define STROKE_RET       6 /* south-west */
                           /* south */
                           /* south-east */

struct {
    char *name;
    int special;
} *specialptr, speciallist[] = {
    {"Shift",    SPECIAL_SHIFT},
    {"Control",  SPECIAL_CONTROL},
    {"Meta",     SPECIAL_META},
    {"Caps",     SPECIAL_CAPS},
    {"Function", SPECIAL_FUNCTION},
    {"=",        0},
    {NULL,       0},
};

static int nstrings;
static int nkeys;
struct keydef *kptr;
struct string *sptr;
int shiftstate;

static void printerror(FILE *ef, char *ehead, char *fname, int lno, char *msg)
{
    if (!ef) return;
    fprintf(ef, "%s%s: %i: %s\n", ehead, fname, lno, msg);
}

static char *parsechar(char *s, int *dest)
{
    while (isspace(*s)) s++;
    if (!isgraph(*s)) {
	*dest = -1;
	return s;
    }
    if (sscanf(s, "0x%x", dest) == 1) {
	while (isgraph(*s)) s++; /* skip it */
    } else {
	/* if not hex, it is a char */
	*dest = *(s++);
    }
    /* skip blanks, again */
    while (isspace(*s)) s++;
    return s;
}

static struct string *lookfor_string(char *name)
{
    struct string *sp;

    if (!sptr) return NULL;
    /* the name can be non-terminated */
    for (sp = sptr; sp->name[0]; sp++)
	if (!strncmp(sp->name, name, strlen(sp->name))
	    && !isgraph(name[strlen(sp->name)]))
	    return sp;
    return NULL;
}

/* returns 0 for success, -errno for error (and message to error-file) */
int kparse_parse_file(char *fname, FILE *ef, char *ehead)
{
    FILE *f;
    char line[128], *s;
    int c, lineno = 0;
    int ret = 0;
    struct keydef *kp;
    struct string *sp, *sp2;

    if (!ehead) ehead = "";
    f = fopen(fname,"r");
    if (!f) {
	if (ef) fprintf(ef, "%s%s: %s\n", ehead, fname, strerror(errno));
	return -errno;
    }

    while (fgets(line, LINE_LEN, f)) {
	lineno++;

	/* ignore long lines */
	if (line[strlen(line)-1] != '\n') {
	    printerror(ef, ehead, fname, lineno, "line too long, skipping");
	    do
		c = getc(f);
	    while (c != '\n' && c != EOF);
	    if (c == EOF) break;
	}

	/* skip blanks */
	for (s = line; *s && isspace(*s); s++)
	    ;

	/* ignore comments and blank lines */
	if (*s == '#' || !*s)
	    continue;
	
	/* manage Nkey and Nstrings */
	if (!strncmp(s, "Nkeys", 5)) {
	    if (nkeys) {
		printerror(ef, ehead, fname, lineno,
			   "ignoring Nkeys redefinition");
		continue;
	    }
	    if (sscanf(s, "Nkeys %i %s", &nkeys, line) != 1) {
		printerror(ef, ehead, fname, lineno, "syntax error");
		ret = -EINVAL;
		continue;
	    }
	    kptr = kp = calloc(sizeof(struct keydef), nkeys+1);
	    kptr[nkeys].special = -1; /* dirty marker */
	    continue;
	}
	if (!strncmp(s, "Nstrings", 8)) {
	    if (nstrings) {
		printerror(ef, ehead, fname, lineno,
			   "ignoring Nstrings redefinition");
		continue;
	    }
	    if (sscanf(s, "Nstrings %i %s", &nstrings, line) != 1) {
		printerror(ef, ehead, fname, lineno, "syntax error");
		ret = -EINVAL;
		continue;
	    }
	    sptr = sp = calloc(sizeof(struct string), nstrings+1);
	    sptr[nstrings].explen = -1; /* dirty marker */
	    continue;
	}

	/* manage definition of a key */
	if (isdigit(*s)) {
	    if (!kptr) {
		printerror(ef, ehead, fname, lineno, "key before Nkeys");
		ret = -EINVAL;
		continue;
	    }
	    if (kp->special == -1) {
		printerror(ef, ehead, fname, lineno, "too many keys");
		ret = -EINVAL;
		continue;
	    }
	    if(sscanf(s, "%d,%d %d,%d %[^\n]", 
		      kp->x,kp->y,kp->x+1,kp->y+1,line) != 5) {
		printerror(ef, ehead, fname, lineno, "syntax error");
		ret = -EINVAL;
		continue;
	    }
	    /* back again: skip blanks */
	    for (s = line; *s && isspace(*s); s++)
		;
	    if (isdigit(*s)) { /* second area (e.g. for the return key) */
		if(sscanf(s, "%d,%d %d,%d %[^\n]", 
			  kp->x+2,kp->y+2,kp->x+3,kp->y+3,line) != 5) {
		    printerror(ef, ehead, fname, lineno, "syntax error");
		    ret = -EINVAL;
		    continue;
		}
		/* back again: skip blanks */
		for (s = line; *s && isspace(*s); s++)
		    ;
	    }

	    /* look for a special name (or an equal sign) */
	    for (specialptr = speciallist; specialptr->name; specialptr++)
		if (!strncmp(specialptr->name, s, strlen(specialptr->name))
		    && !isgraph(s[strlen(specialptr->name)]))
		    break;
	    if (specialptr->special) {
		/* found; it's a special key */
		(kp++)->special = specialptr->special;
		continue;
	    }
	    if (specialptr->name) {
		/* found: it's an equal sign. Now parse 1 or 2 characters */
		s++;
		s = parsechar(s, kp->codes);
		if (*s) {
		    s = parsechar(s, kp->codes+1);
		} else {
		    kp->codes[1] = kp->codes[0];
		    if (islower(kp->codes[1]))
			kp->codes[1] = toupper(kp->codes[1]);
		}
		if (*s) {
		    printerror(ef, ehead, fname, lineno, "too much data");
		    ret = -EINVAL;
		    continue;
		}
		if (kp->codes[1] == -1) kp->codes[1] = kp->codes[0];
		if (kp->codes[0] == -1) {
		    printerror(ef, ehead, fname, lineno, "syntax error");
		    ret = -EINVAL;
		    continue;
		}
		kp++;
		continue;
	    }

	    /* not found: it must be a string */
	    sp2 = lookfor_string(s);
	    if (!sp2) {
		printerror(ef, ehead, fname, lineno, "syntax error");
		ret = -EINVAL;
		continue;
	    }
	    kp++->sptr = sp2;
	    continue;
	} /* isdigit(*s) */

	/* manage definition of a string */
	if (!strncmp(s, "String", 6)) {
	    if (!sptr) {
		printerror(ef, ehead, fname, lineno, "string before Nstrings");
		ret = -EINVAL;
	    }
	    if (sp->explen == -1) {
		printerror(ef, ehead, fname, lineno, "too many strings");
		ret = -EINVAL;
		continue;
	    }
	    if (sscanf(s, "String %s %[^\n]", sp->name, line) != 2) {
		printerror(ef, ehead, fname, lineno, "syntax error");
		ret = -EINVAL;
		continue;
	    }
	    s = line;
	    for (sp->explen = 0;
		 *s && sp->explen < STRINGEXP_LEN;
		 sp->explen++) {
		s = parsechar(s, &c);
		sp->expand[sp->explen] = c;
		if (c==-1) {
		    printerror(ef, ehead, fname, lineno, "syntax error");
		    ret = -EINVAL;
		    continue;
		}
	    }
	    sp++;
	    continue;
	}

	/* nothing known */
	printerror(ef, ehead, fname, lineno, "syntax error");
	ret = -EINVAL;
    } /* while fgets() */

    fclose(f);
    return ret;
}

static void toggle_special(int special, fbinfo *devinfo)
{
    struct keydef *kp;

    if (!kptr) return;

    shiftstate ^= special;

    if (special == SPECIAL_SHIFT) {
      // redraw the keyboard in the new shifted state.
      draw_keyboard(devinfo);
    }

    for (kp = kptr; kp->special >= 0; kp++)
      if (kp->special == special && 
	  (kp->special != SPECIAL_SHIFT || shiftstate & SPECIAL_SHIFT) &&
	  (kp->special != SPECIAL_CAPS || shiftstate & SPECIAL_CAPS))
	kdraw_xor(devinfo->offscreenkeymap, kp->x, kp->y, devinfo);
}



/* returns -EAGAIN if more data is there */
int kparse_usestroke(int coords[4], char *result, int *rlen, fbinfo *devinfo)
{
    struct keydef *kp;
    struct string *sp;
    int c;
    int strokedir = STROKE_NONE;

#ifdef USE_STROKE_TOO

    int dx, dy, squarelen;

    dx = coords[2]-coords[0];
    dy = coords[3]-coords[1];
    squarelen = dx*dx + dy*dy;

    /* use height of first key as a reference, must drag for .7 height */
    if (squarelen*2 > (kptr->y[1]-kptr->y[0])*(kptr->y[1]-kptr->y[0])) {
#if 0 /* no, don't use eight directions */
#if 0 /* it looks like I have no libm in my ramdisk */
	/* the direction is 1 for east, each is 45deg wide */
	ddir = atan2((double)dy, (double)dx);
	ddir += M_PI/8 + 2*M_PI;
	strokedir = (int)(ddir / (M_PI/4)) % 8 + 1; /* range is 1-8 */
#else /* dirty hack to do it by hand */
	if (dx == 0) {
	    if (dy>0) strokedir = 3; /* north */
	    else strokedir = 7; /* south */
	} else {
	    ddir = (double)-dy/dx; /* I think pithagoric. so -dx */
	    strokedir = 7;
	    if (ddir > -2.414) strokedir = 8; /* more than -77.5 */
	    if (ddir > -0.414) strokedir = 1; /* more than -22.5 */
	    if (ddir >  0.414) strokedir = 2; /* more than  22.5 */
	    if (ddir >  2.414) strokedir = 3; /* more than  77.5 */
	    if (dx <0) strokedir += 4;
	    strokedir = (strokedir -1) % 8 + 1;
	}
#endif
#else /* use four directions */
	if (abs(dx) > abs(dy)) {
	    if (dx > 0) strokedir = STROKE_SPACE;
	    else        strokedir = STROKE_DEL;
	} else {
	    if (dy > 0) strokedir = STROKE_RET;
	    else        strokedir = STROKE_SHIFT;
	}
#endif
	fprintf(stderr, "dx %3i dy %3i strokedir %i\n", dx, dy, strokedir);
    }

#endif

    //    printf("Parsing coords %d x %d\n", coords[0], coords[1]);

    *rlen = 0;
    for (kp = kptr; kp->special >= 0; kp++) {
	/* first rectangle */
	if (coords[0] >= kp->x[0] && coords[0] <= kp->x[1]
	    && coords[1] >= kp->y[0] && coords[1] <= kp->y[1])
	    break;
	/* second rectangle */
	if (coords[0] >= kp->x[2] && coords[0] <= kp->x[3]
	    && coords[1] >= kp->y[2] && coords[1] <= kp->y[3])
	    break;
    }
    if (kp->special < 0) {
	/* not found */
	*rlen = 0;
	return 0;
    }

    /* ok, now we know the landing point. Use it */
    if (kp->special) {
	toggle_special(kp->special, devinfo);
	return 0;
    }

    /* flash the key */
    kdraw_xor(devinfo->addr, kp->x, kp->y, devinfo); usleep(100 * 1000); kdraw_xor(devinfo->addr, kp->x, kp->y, devinfo);

    if (kp->sptr) { /* strings do not depend on shiftstate */
	*rlen = kp->sptr->explen; /* FIXME: check lenght */
	memcpy(result, kp->sptr->expand, kp->sptr->explen);
	return 0;
    }
    c = kp->codes[0];

    /* The function shift key is only used with digits (What about 11-12?) */
    if (!isdigit(c) && (shiftstate & SPECIAL_FUNCTION)) {
	toggle_special(SPECIAL_FUNCTION, devinfo);
    }

    if (shiftstate & SPECIAL_FUNCTION) {
	/* we are sure this is a digit */
	char fstring[4];
	int f = c - '0' ?: 10; /* '0' is F10 */

	toggle_special(SPECIAL_FUNCTION, devinfo);
	if (shiftstate & SPECIAL_SHIFT) {
	    toggle_special(SPECIAL_SHIFT, devinfo);
	    f += 10;
	}
	if (shiftstate & SPECIAL_META) {
	    toggle_special(SPECIAL_META, devinfo);
	    /* need to switch console: return it as a special marker */
	    *rlen = -1;
	    result[0] = f;
	    return 0;
	}
	sprintf(fstring, "F%i",f);
	sp = lookfor_string(fstring);
	if (!sp) {
	    *rlen = 0; return 0;
	}
	*rlen = sp->explen;
	memcpy(result, sp->expand, sp->explen);
	return 0;
    }	

    /* This is a normal key, modify it according to shiftstate */
    if (shiftstate & SPECIAL_SHIFT) {
	toggle_special(SPECIAL_SHIFT, devinfo);
	c = kp->codes[1];
    }
    if (strokedir == STROKE_SHIFT) {
	c = kp->codes[1];
    }	
    if (shiftstate & SPECIAL_META) {
	toggle_special(SPECIAL_META, devinfo);
	c += 128;
    }
    if (shiftstate & SPECIAL_CONTROL) {
	toggle_special(SPECIAL_CONTROL, devinfo);
	if ((c >= 0x40 && c < 0x60) || islower(c))
	    c &= 0x1f;
    }
    if (shiftstate & SPECIAL_CAPS && isalpha(c)) {
	c ^= 0x20;
    }
    *rlen = 1;
    result[0] = c;

#ifdef USE_STROKE_TOO

    switch(strokedir) {
        case STROKE_SPACE:
	    result[(*rlen)++] = ' ';
	    break;
        case STROKE_DEL:
	    result[0] = 0x7f; *rlen = 1;
	    break;
        case STROKE_RET:
	    result[(*rlen)++] = '\n';
	    break;
	 /* STROKE_SHIFT already handled */
    }
#endif

    return 0;
}
	

	    
#ifdef MAIN /* extra stuff to check that it works (used by "testtool") */

void kparse_dump_structs(void)
{
    struct keydef *kp;
    struct string *sp;

    for (kp = kptr; kp->special >= 0; kp++) {
	printf("%i,%i %i,%i %i,%i %i,%i 0x%02x (%c) 0x%02x (%c) \"%s\" %i\n",
	       kp->x[0],kp->y[0],
	       kp->x[1],kp->y[1],
	       kp->x[2],kp->y[2],
	       kp->x[3],kp->y[3],
	       kp->codes[0], kp->codes[0] > 0x1f ? kp->codes[0] : ' ',
	       kp->codes[1], kp->codes[1] > 0x1f ? kp->codes[1] : ' ',
	       kp->sptr ? kp->sptr->expand : "",
	       kp->special);
    }
    for (sp = sptr; sp->explen > 0; sp++) {
	printf("string %s %i:\"%s\"\n", sp->name, sp->explen, sp->expand);
    }
}

char *prgname;

int main(int argc, char **argv)
{
    char s[128];
    char res[10];
    int i, c[4], rlen;
    fbinfo *devinfo;

    printf("This isn't done, and won't work. The code needs finishing.\n");
    exit(1);

    prgname = argv[0];
    if (kparse_parse_file(argv[1], stderr, "kparsetest: ")) exit(1);
    /* kparse_dump_structs(); */
    while (fgets(s,128,stdin)) {
	if (sscanf(s, "%i %i %i %i",c, c+1, c+2, c+3) != 4) {
	    fprintf(stderr, "invalid data: %s",s);
	    continue;
	}
	kparse_usestroke(c, res, &rlen, devinfo);
	if (rlen == -1) {
	    printf("should switch to console %i\n", res[0]);
	    continue;
	}
	for (i=0; i<rlen; i++)
	    printf( "'%c' (0x%02x)\n", res[i] > 0x1f ? res[i] : '?', res[i]);
    }		    
    return(0);
}

#endif /* MAIN */

