/*
 * mwin.c -- routines for touch panel calibration and multiple bitdepth
 *           drawing, based on (or copied from) MicroWindows.
 *
 * -- Jorj Bauer <jorj@seas.upenn.edu>
 * sponsored by DejaVu Software, Inc. (http://www.dejavusoftware.com/).
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>

#include "key-engine.h"

static int enable_pointing_coordinate_transform = 1;
TRANSFORMATION_COEFFICIENTS tc;
extern int pd_fd;

inline MWPOINT DeviceToScreen(MWPOINT p)
{
	/*
	 * Transform device coordinates to screen coordinates.
	 * Take a point p in device coordinates and return the corresponding
	 * point in screen coodinates.
	 * This can scale, translate, rotate and/or skew, based on the
	 * coefficients calculated above based on the list of screen
	 * vs. device coordinates.
	 */

	static MWPOINT prev;
	/* set slop at 3/4 pixel */
	const short slop = TRANSFORMATION_UNITS_PER_PIXEL * 3 / 4;
	MWPOINT new, out;

	/* transform */
	new.x = (tc.a * p.x + tc.b * p.y + tc.c) / tc.s;
	new.y = (tc.d * p.x + tc.e * p.y + tc.f) / tc.s;

	/* hysteresis (thanks to John Siau) */
	if ( abs(new.x - prev.x) >= slop )
		out.x = (new.x | 0x3) ^ 0x3;
	else
		out.x = prev.x;

	if ( abs(new.y - prev.y) >= slop )
		out.y = (new.y | 0x3) ^ 0x3;
	else
		out.y = prev.y;

	prev = out;

	return out;
}

int PD_Read(MWCOORD *px, MWCOORD *py, MWCOORD *pz, int *pb)
{
	/*
	 * Read the tpanel state and position.
         * Returns the position data in x, y, and button data in b.
	 * Returns -1 on error.
	 * Returns 0 if no new data is available.
	 * Returns 1 if position data is relative (i.e. mice).
	 * Returns 2 if position data is absolute (i.e. touch panels).
	 * Returns 3 if position data is not available, but button data is.
	 * This routine does not block.
	 *
	 * Unlike a mouse, this driver returns absolute postions, not deltas.
	 */

	/* If z is below this value, ignore the data. */
	/* const int low_z_limit = 900; EVEREX*/
	const int low_z_limit = 815;

	/*
	 * I do some error masking by tossing out really wild data points.
	 * Lower data_change_limit value means pointer get's "left behind"
	 * more easily.  Higher value means less errors caught.
	 * The right setting of this value is just slightly higher than
	 * the number of units traversed per sample during a "quick" stroke.
	 */
	const int data_change_limit = DATACHANGELIMIT;
	static int have_last_data = 0;
	static int last_data_x = 0;
	static int last_data_y = 0;

	/*
	 * Thanks to John Siau <jsiau@benchmarkmedia.com> for help with the
	 * noise filter.  I use an infinite impulse response low-pass filter
	 * on the data to filter out high-frequency noise.  Results look
	 * better than a finite impulse response filter.
	 * If I understand it right, the nice thing is that the noise now
	 * acts as a dither signal that effectively increases the resolution
	 * of the a/d converter by a few bits and drops the noise level by
	 * about 10db.
	 * Please don't quote me on those db numbers however. :-)
	 * The end result is that the pointer goes from very fuzzy to much
	 * more steady.  Hysteresis really calms it down in the end (elsewhere).
	 *
	 * iir_shift_bits effectively sets the number of samples used by
	 * the filter * (number of samples is 2^iir_shift_bits).
	 * Lower iir_width means less pointer lag, higher iir_width means
	 * steadier pointer.
	 */
	const int iir_shift_bits = 3;
	const int iir_sample_depth = (1 << iir_shift_bits);
	static int iir_accum_x = 0;
	static int iir_accum_y = 0;
	static int iir_accum_z = 0;
	static int iir_count = 0;

	int data_x, data_y, data_z;

	/* read a data point */
	short data[6];
	int bytes_read;
	bytes_read = read(pd_fd, data, sizeof(data));
	if (bytes_read != sizeof(data)) {
		if (errno == EINTR || errno == EAGAIN) {
			return 0;
		}
		return -1;
	}

	/* did we lose any data? */
	if ( (data[0] & 0x2000) )
		EPRINTF("Lost touch panel data\n");

	/* do we only have contact state data (no position data)? */
	if ( (data[0] & 0x8000) == 0 )
	{
		/* is it a pen-release? */
		if ( (data[0] & 0x4000) == 0 )
		{
			/* reset the limiter */
			have_last_data = 0;

			/* reset the filter */
			iir_count = 0;
			iir_accum_x = 0;
			iir_accum_y = 0;
			iir_accum_z = 0;

			/* return the pen (button) state only, */
			/* indicating that the pen is up (no buttons are down)*/
			*pb = 0;
			return 3;
		}

		/* ignore pen-down since we don't know where it is */
		return 0;
	}

	/* we have position data */

	/*
	 * Combine the complementary panel readings into one value (except z)
	 * This effectively doubles the sampling freqency, reducing noise
	 * by approx 3db.
	 * Again, please don't quote the 3db figure.  I think it also
	 * cancels out changes in the overall resistance of the panel
	 * such as may be caused by changes in panel temperature.
	 */
	data_x = data[2] - data[1];
	data_y = data[4] - data[3];
	data_z = data[5];

	/* isn't z big enough for valid position data? */
	if ( data_z <= low_z_limit ) {
		return 0;
	}

	/* has the position changed more than we will allow? */
	if ( have_last_data )
		if ( (abs(data_x - last_data_x) > data_change_limit)
			|| ( abs(data_y - last_data_y) > data_change_limit ) ) {
			return 0;
		}

	/* save last position */
	last_data_x = data_x;
	last_data_y = data_y;
	have_last_data = 1;

	/* is filter ready? */
	if ( iir_count == iir_sample_depth )
	{
		/* make room for new sample */
		iir_accum_x -= iir_accum_x >> iir_shift_bits;
		iir_accum_y -= iir_accum_y >> iir_shift_bits;
		iir_accum_z -= iir_accum_z >> iir_shift_bits;

		/* feed new sample to filter */
		iir_accum_x += data_x;
		iir_accum_y += data_y;
		iir_accum_z += data_z;

		/* transformation enabled? */
		if (enable_pointing_coordinate_transform)
		{
			/* transform x,y to screen coords */
			MWPOINT transformed = {iir_accum_x, iir_accum_y};
			transformed = DeviceToScreen(transformed);
			/*
			 * HACK: move this from quarter pixels to whole
			 * pixels for now at least until I decide on the
			 * right interface to get the quarter-pixel data
			 * up to the next layer.
			 */
			*px = transformed.x >> 2;
			*py = transformed.y >> 2;
		}
		else
		{
			/* return untransformed coords (for calibration) */
			*px = iir_accum_x;
			*py = iir_accum_y;
		}

		/* return filtered pressure */
		*pb = MWBUTTON_L;
		*pz = iir_accum_z;

#ifdef TEST
		EPRINTF("In: %hd, %hd, %hd  Filtered: %d %d %d  Out: %d, %d, %d\n",
			data_x, data_y, data_z, iir_accum_x, iir_accum_y, iir_accum_z, *px, *py, *pz);
#endif
		return 2;
	}

	/* prime the filter */
	iir_accum_x += data_x;
	iir_accum_y += data_y;
	iir_accum_z += data_z;
	iir_count += 1;

	return 0;
}


/* We need depth-dependant functions for drawing hlines, vlines, dots, and
   doing xors. These functions are rewrites based on the MicroWindows 
   source. -- jorj, 7/4/00 */

/* 1bpp drawing functions. */

static unsigned char notmask_1[8] = {
	0x7f, 0xbf, 0xdf, 0xef, 0xf7, 0xfb, 0xfd, 0xfe};

void linear1_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, int c)
{
  unsigned char *addr = buffer + (x>>3) + y * devinfo->linelen;
  *addr = (*addr & notmask_1[x&7]) | (c << (7-(x&7)));
}

void linear1_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		       int c, int mode)
{
  unsigned char *addr = buffer + (x1>>3) + y * devinfo->linelen;
  while (x1 <= x2) {
    if (mode == MODE_XOR)
      *addr ^= c << (7-(x1&7));
    else
      *addr = (*addr & notmask_1[x1&7]) | (c << (7-(x1&7)));
    if ((++x1 & 7) == 0)
      ++addr;
  }
}

void linear1_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
		       int c)
{
  unsigned char *addr = buffer + (x>>3) + y1 * devinfo->linelen;

  while (y1++ <= y2) {
    *addr = (*addr & notmask_1[x&7]) | (c << (7-(x&7)));
    addr += devinfo->linelen;
  }
}


/* 2bpp drawing functions. */

static unsigned char notmask_2[4] = { 0x3f, 0xcf, 0xf3, 0xfc };

void linear2_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, int c)
{
  unsigned char *addr = buffer + (x>>2) + y*devinfo->linelen;
  *addr = (*addr & notmask_2[x&3]) | (c << ((3-(x&3))<<1));
}

void linear2_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		       int c, int mode)
{
  unsigned char *addr = buffer + (x1>>2) + y * devinfo->linelen;
  while (x1 <= x2) {
    if (mode == MODE_XOR)
      *addr ^= c << ((3-(x1&3))<<1);
    else
      *addr = (*addr & notmask_2[x1&3]) | (c << ((3-(x1&3))<<1));
    if ((++x1 & 3) == 0)
      ++addr;
  }
}

void linear2_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
		       int c)
{
  unsigned char *addr = buffer + (x>>2) + y1 * devinfo->linelen;

  while (y1++ <= y2) {
    *addr = (*addr & notmask_2[x&3]) | (c << ((3-(x&3))<<1));
    addr += devinfo->linelen;
  }
}

/* 4bpp drawing functions. */

static unsigned char notmask_4[2] = { 0x0f, 0xf0 };

void linear4_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, int c)
{
  unsigned char *addr;
  
  addr = buffer + (x>>1) + y*devinfo->linelen;
  *addr = (*addr & notmask_4[x&1]) | (c << ((1-(x&1))<<2));
}

void linear4_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		       int c, int mode)
{
  unsigned char *addr;

  addr = buffer + (x1>>1) + y*devinfo->linelen;
  while (x1 <= x2) {
    if (mode == MODE_XOR)
      *addr ^= c << ((1-(x1&1))<<2);
    else
      *addr = (*addr & notmask_4[x1&1]) | (c << ((1-(x1&1))<<2));
    if ((++x1 & 1) == 0)
      ++addr;
  }
}

void linear4_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
		       int c)
{
  unsigned char *addr;

  addr = buffer + (x>>1) + y1 * devinfo->linelen;
  while (y1++ <= y2) {
    *addr = (*addr & notmask_4[x&1]) | (c << ((1-(x&1))<<2));
    addr += devinfo->linelen;
  }
}

/* 8bpp drawing functions. */

void linear8_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, int c)
{
  __u8 *addr = buffer;

  addr[x + y * devinfo->linelen] = c;
}

void linear8_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		       int c, int mode)
{
  unsigned char *addr = buffer + x1 + y * devinfo->linelen;
  if (mode == MODE_XOR) {
    while (x1++ <= x2)
      *addr++ ^= c;
  } else {
    memset(addr, c, x2 - x1 + 1);
  }
}

void linear8_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
		       int c)
{
  unsigned char *addr = buffer + x + y1 * devinfo->linelen;
  while (y1++ <= y2) {
    *addr = c;
    addr += devinfo->linelen;
  }
}

/* 16bpp drawing functions. */

void linear16_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, __u16 c)
{
  __u16 *addr = (__u16 *)buffer;
  addr[x + y * (devinfo->linelen>>1)] = c;
}

void linear16_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		       __u16 c, int mode)
{
  __u16 *addr = (__u16 *)buffer + x1 + y * (devinfo->linelen>>1);

  while (x1++ <= x2)
    if (mode == MODE_XOR)
      *addr++ ^= c;
    else *addr++ = c;
}

void linear16_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
		       __u16 c)
{
  __u16 *addr = (__u16 *)buffer + x + y1 * (devinfo->linelen>>1);
  
  while (y1++ <= y2) {
    *addr = c;
    addr += devinfo->linelen>>1;
  }
}

/* 24bpp drawing functions. */

/* return 8/8/8 bit r, g or b component of 24 bit pixelval*/
#define PIXEL888RED(pixelval)		(((pixelval) >> 16) & 0xff)
#define PIXEL888GREEN(pixelval)		(((pixelval) >> 8) & 0xff)
#define PIXEL888BLUE(pixelval)		((pixelval) & 0xff)

void linear24_drawpixel(char *buffer, fbinfo *devinfo, int x, int y, 
			__u32 c)
{
  __u8 *addr = buffer;
  unsigned char r, g, b;

  r = PIXEL888RED(c);
  g = PIXEL888GREEN(c);
  b = PIXEL888BLUE(c);

  addr += x + y * devinfo->linelen;
  *addr++ = b;
  *addr++ = g;
  *addr = r;
 
}

void linear24_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
			__u32 c, int mode)
{
  unsigned char *addr = buffer;
  unsigned char r, g, b;

  r = PIXEL888RED(c);
  g = PIXEL888GREEN(c);
  b = PIXEL888BLUE(c);
  addr += x1 + y * devinfo->linelen;

  while (x1++ <= x2) {
    if (mode == MODE_XOR) {
      *addr++ ^= b;
      *addr++ ^= g;
      *addr++ ^= r;
    } else {
      *addr++ = b;
      *addr++ = g;
      *addr++ = r;
    }
  }
}

void linear24_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
			__u32 c)
{
  unsigned char *addr = buffer;
  unsigned char r, g, b;

  r = PIXEL888RED(c);
  g = PIXEL888GREEN(c);
  b = PIXEL888BLUE(c);
  addr += x + y1 * devinfo->linelen;
  
  while (y1++ <= y2) {
    addr[0] = b;
    addr[1] = g;
    addr[2] = r;
    addr += devinfo->linelen/3;
  }
}

/* 32bpp drawing functions. */

void linear32_drawpixel(char *buffer, fbinfo *devinfo, int x, int y,
			__u32 c)
{
  __u32 *addr = (__u32 *)buffer;

  addr[x + y * (devinfo->linelen>>2)] = c;
}

void linear32_drawhline(char *buffer, fbinfo *devinfo, int x1, int x2, int y,
		        __u32 c, int mode)
{
  __u32 *addr = (__u32 *)buffer + x1 + y * (devinfo->linelen>>2);

  while (x1 ++ <= x2)
    if (mode == MODE_XOR)
      *addr++ ^= c;
    else *addr++ = c;

}

void linear32_drawvline(char *buffer, fbinfo *devinfo, int y1, int y2, int x,
			__u32 c)
{
  __u32 *addr = (__u32 *)buffer + x + y1 * (devinfo->linelen>>2);
  while (y1++ <= y2) {
    *addr = c;
    addr += devinfo->linelen>>2;
  }
}
